from dataclasses import dataclass

from omni_converter.solver.rules import AutoRuleBook, CastLambda, ConversionLambda, RuleEdge
from omni_cv_rules.coconut.convert import imdef_neighbors, rule_xyz_to_rgb, rule_batch_xyz_to_rgb, \
    rule_VR_None_to_normalized, rule_add_channel, rule_swap_RGB_BGR
from omni_cv_rules.coconut.omni_converter import cast_imdef_str_to_imdef, cast_imdef_to_imdef_str, dict2imdef, \
    cast_ary_str_to_ary_type, img_list_is_imgs, lll_is_rgb, cast_tuple2auto_tuple, rule_format, rule_imgs2tile, \
    rule_batch_to_tiled, rule_img2widget, rule_numpy2img, rule_image2gray, rule_image_path_to_image, numpys_to_numpy, \
    tensor_to_list, rgb_to_rgba, repeat_ch, torch_img_to_pixpix_input, torch_img_to_vgg_prep, rule_to_spectrum, \
    cast_imdef_to_dict, rule_image2lab, rule_lab_value_conversion
from omni_cv_rules.coconut.recursive import intra_list_conversions
from omni_cv_rules.coconut.to_bytes import rule_image_str_to_jpg_bytes
from omni_cv_rules.coconut.widgets import any2widget, auto_tuple2widget
from omni_cv_rules.custom_rules import custom_rule_book


def create_cast_rule(rule, name=None, cost=1):
    """
    rule: State->List[State] # should return list of possible casts without data conversion.
    """
    return CastLambda(rule, name, cost=cost)


def create_conversion_rule(rule):
    return ConversionLambda(rule)


def create_alias_rule(a, b):
    def caster(state):
        if state == a:
            return [b]
        elif state == b:
            return [a]

    return create_cast_rule(caster, f"alias:{a}=={b}")


def rule_to_hsv(state):
    if state == "image,RGB,RGB":
        return [RuleEdge(
            converter=lambda img: img.convert("HSV"),
            new_format="image,HSV,HSV",
            cost=1,
            name=f"PIL RGB to PIL HSV"
        )]
    elif state == "image,HSV,HSV":
        return [RuleEdge(
            converter=lambda img: img.convert("RGB"),
            new_format="image,RGB,RGB",
            cost=1,
            name=f"PIL HSV to PIL RGB"
        )]


CV_RULEBOOK = (AutoRuleBook(id="CV_RULEBOOK").add_rules(
    imdef_neighbors,
    rule_xyz_to_rgb,
    rule_batch_xyz_to_rgb,
    rule_VR_None_to_normalized,
    rule_add_channel,
    rule_swap_RGB_BGR,
    create_cast_rule(cast_imdef_to_dict, "cast_imdef_to_dict"),
    create_cast_rule(cast_imdef_str_to_imdef, "cast_imdef_str_to_imdef"),
    create_cast_rule(cast_imdef_to_imdef_str, "cast_imdef_to_imdef_str"),
    create_cast_rule(dict2imdef, "dict2imdef"),
    create_cast_rule(cast_ary_str_to_ary_type, "cast_ary_str_to_ary_type"),
    create_cast_rule(img_list_is_imgs, "img_list_is_imgs"),
    create_cast_rule(lll_is_rgb, "lll_is_rgb", cost=10),
    create_cast_rule(cast_tuple2auto_tuple, "tuple <--> auto_tuple"),
    # AutoSolver.create_cast_rule(dict2visdomable),
    create_conversion_rule(any2widget),
    create_conversion_rule(rule_format),
    create_conversion_rule(rule_imgs2tile),
    create_conversion_rule(rule_batch_to_tiled),
    create_conversion_rule(rule_img2widget),
    create_conversion_rule(rule_numpy2img),
    create_conversion_rule(rule_image2gray),
    create_conversion_rule(rule_image_path_to_image),
    create_conversion_rule(rule_image_str_to_jpg_bytes),
    create_conversion_rule(rule_image2lab),
    # AutoSolver.create_conversion_rule(rule_rgba2laba),
    create_conversion_rule(rule_lab_value_conversion),
    create_conversion_rule(numpys_to_numpy),
    create_conversion_rule(tensor_to_list),
    create_conversion_rule(rule_to_hsv),
    create_conversion_rule(rgb_to_rgba),
    create_conversion_rule(repeat_ch),
    create_conversion_rule(torch_img_to_pixpix_input),
    create_conversion_rule(torch_img_to_vgg_prep),
    # AutoSolver.create_conversion_rule(intra_tuple_conversions),
    create_conversion_rule(rule_to_spectrum),
    create_conversion_rule(auto_tuple2widget),
    create_alias_rule("numpy_rgb", "numpy,uint8,HWC,RGB,0_255"),
    create_alias_rule("numpy_rgba", "numpy,uint8,HWC,RGBA,0_255"),
) + custom_rule_book.add_recursive_rule(
    intra_list_conversions
)).set_id("CV_RULEBOOK")



class RealTorchProxy:
    @property
    def torch(self):
        import torch
        return torch

    def from_numpy(self, ary):
        return self.torch.from_numpy(ary)
    def cat(self,ary,dim):
        return self.torch.cat(ary,dim)
    def tensor(self,ary):
        return self.torch.tensor(ary)

torch = RealTorchProxy()





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# __coconut_hash__ = 0x3aa8491

# Compiled with Coconut version 1.6.0 [Vocational Guidance Counsellor]

# Coconut Header: -------------------------------------------------------------

from __future__ import generator_stop
import sys as _coconut_sys, os as _coconut_os
_coconut_file_dir = _coconut_os.path.dirname(_coconut_os.path.abspath(__file__))
_coconut_cached_module = _coconut_sys.modules.get("__coconut__")
if _coconut_cached_module is not None and _coconut_os.path.dirname(_coconut_cached_module.__file__) != _coconut_file_dir:
    del _coconut_sys.modules["__coconut__"]
_coconut_sys.path.insert(0, _coconut_file_dir)
_coconut_module_name = _coconut_os.path.splitext(_coconut_os.path.basename(_coconut_file_dir))[0]
if _coconut_module_name and _coconut_module_name[0].isalpha() and all(c.isalpha() or c.isdigit() for c in _coconut_module_name) and "__init__.py" in _coconut_os.listdir(_coconut_file_dir):
    _coconut_full_module_name = str(_coconut_module_name + ".__coconut__")
    import __coconut__ as _coconut__coconut__
    _coconut__coconut__.__name__ = _coconut_full_module_name
    for _coconut_v in vars(_coconut__coconut__).values():
        if getattr(_coconut_v, "__module__", None) == "__coconut__":
            try:
                _coconut_v.__module__ = _coconut_full_module_name
            except AttributeError:
                _coconut_v_type = type(_coconut_v)
                if getattr(_coconut_v_type, "__module__", None) == "__coconut__":
                    _coconut_v_type.__module__ = _coconut_full_module_name
    _coconut_sys.modules[_coconut_full_module_name] = _coconut__coconut__
from __coconut__ import *
from __coconut__ import _coconut_call_set_names, _coconut, _coconut_MatchError, _coconut_igetitem, _coconut_base_compose, _coconut_forward_compose, _coconut_back_compose, _coconut_forward_star_compose, _coconut_back_star_compose, _coconut_forward_dubstar_compose, _coconut_back_dubstar_compose, _coconut_pipe, _coconut_star_pipe, _coconut_dubstar_pipe, _coconut_back_pipe, _coconut_back_star_pipe, _coconut_back_dubstar_pipe, _coconut_none_pipe, _coconut_none_star_pipe, _coconut_none_dubstar_pipe, _coconut_bool_and, _coconut_bool_or, _coconut_none_coalesce, _coconut_minus, _coconut_map, _coconut_partial, _coconut_get_function_match_error, _coconut_base_pattern_func, _coconut_addpattern, _coconut_sentinel, _coconut_assert, _coconut_mark_as_match, _coconut_reiterable, _coconut_self_match_types, _coconut_dict_merge, _coconut_exec
_coconut_sys.path.pop(0)

# Compiled Coconut: -----------------------------------------------------------

import ipywidgets as widgets  # import ipywidgets as widgets
from pprintpp import pformat  # from pprintpp import pformat,pprint
from pprintpp import pprint  # from pprintpp import pformat,pprint
from PIL import Image  # from PIL import Image
import PIL  # import PIL
from loguru import logger  # from loguru import logger
import numpy as np  # import numpy as np
import pandas as pd  # import pandas as pd
from typing import Iterable  # from typing import Iterable
nn = lambda none, func: func(none) if none is not None else None  # nn = (none,func) -> func(none) if none is not None else None
L = _coconut.functools.partial(Image.fromarray, mode="L")  # L = Image.fromarray$(mode="L")
RGB = _coconut.functools.partial(Image.fromarray, mode="RGB")  # RGB = Image.fromarray$(mode="RGB")
RGBA = _coconut.functools.partial(Image.fromarray, mode="RGBA")  # RGBA = Image.fromarray$(mode="RGBA")
batch_image_to_image = lambda ary: ary.transpose((1, 0, 2, 3)).reshape((ary.shape[2], -1, ary.shape[3]))  # batch_image_to_image = ary->ary.transpose((1, 0, 2, 3)).reshape((ary.shape[2], -1, ary.shape[3]))
batch_L_to_img = lambda ary: ary  # batch_L_to_img = ary->ary
def img_to_widget(value):  # def img_to_widget(value):
    img_widget = widgets.Image(value=value._repr_png_(), format="png")  #     img_widget = widgets.Image(value=value._repr_png_(),format="png")
    img_widget.layout.object_fit = "contain"  #     img_widget.layout.object_fit = "contain"
    return (img_widget)  #     return img_widget
#img_to_widget = value ->widgets.Box([widgets.Image(value=value._repr_png_(),format="png")])

from bqplot import pyplot as blt  # from bqplot import pyplot as blt
from loguru import logger  # from loguru import logger

class DummyType(_coconut.collections.namedtuple("DummyType", ())):  # data DummyType
    __slots__ = ()  # data DummyType
    __ne__ = _coconut.object.__ne__  # data DummyType
    def __eq__(self, other):  # data DummyType
        return self.__class__ is other.__class__ and _coconut.tuple.__eq__(self, other)  # data DummyType
    def __hash__(self):  # data DummyType
        return _coconut.tuple.__hash__(self) ^ hash(self.__class__)  # data DummyType
    __match_args__ = ()  # data DummyType

_coconut_call_set_names(DummyType)  # try:
try:  # try:
    import torch  #     import torch
    TensorType = torch.Tensor  #     TensorType = torch.Tensor
except Exception as e:  # except Exception as e:
    logger.warning("failed to load torch. no visualization for torch tensor.".format())  #     logger.warning(f"failed to load torch. no visualization for torch tensor.")
    TensorType = DummyType  #     TensorType = DummyType

def bqplot_hist(ary, title="histogram", bins=50):  # def bqplot_hist(ary, title="histogram", bins=50):
    fig = blt.figure(title=title)  #     fig = blt.figure(title=title)
    blt.hist(ary.flatten(), bins=bins)  #     blt.hist(ary.flatten(), bins=bins)
    fig.layout.width = "400px"  #     fig.layout.width = "400px"
    fig.layout.height = "300px"  #     fig.layout.height = "300px"
    return (fig)  #     return fig


def ary_to_image(value):  # def ary_to_image(value):
    if not (isinstance)(value, np.ndarray):  #     if not value `isinstance` np.ndarray:
        return (None)  #         return None
    if value.dtype is not np.uint8:  #     if value.dtype is not np.uint8:
        if value.max() <= 1:  #         if value.max() <= 1:
            value = value * 255  #             value = value * 255
#logger.warning("automatically converting dtype {value.dtype} to uint8 for visualization")
        value = value.astype("uint8")  #         value = value.astype("uint8")


    _coconut_case_match_to_0 = value.shape  #     case value.shape:
    _coconut_case_match_check_0 = False  #     case value.shape:
    _coconut_match_set_name_bs = _coconut_sentinel  #     case value.shape:
    _coconut_match_set_name_h = _coconut_sentinel  #     case value.shape:
    _coconut_match_set_name_w = _coconut_sentinel  #     case value.shape:
    if (_coconut.isinstance(_coconut_case_match_to_0, _coconut.abc.Sequence)) and (_coconut.len(_coconut_case_match_to_0) == 4) and (_coconut_case_match_to_0[3] == 1):  #     case value.shape:
        _coconut_match_set_name_bs = _coconut_case_match_to_0[0]  #     case value.shape:
        _coconut_match_set_name_h = _coconut_case_match_to_0[1]  #     case value.shape:
        _coconut_match_set_name_w = _coconut_case_match_to_0[2]  #     case value.shape:
        _coconut_case_match_check_0 = True  #     case value.shape:
    if _coconut_case_match_check_0:  #     case value.shape:
        if _coconut_match_set_name_bs is not _coconut_sentinel:  #     case value.shape:
            bs = _coconut_case_match_to_0[0]  #     case value.shape:
        if _coconut_match_set_name_h is not _coconut_sentinel:  #     case value.shape:
            h = _coconut_case_match_to_0[1]  #     case value.shape:
        if _coconut_match_set_name_w is not _coconut_sentinel:  #     case value.shape:
            w = _coconut_case_match_to_0[2]  #     case value.shape:
    if _coconut_case_match_check_0:  #     case value.shape:
        return ((L)(value.transpose(1, 0, 2, 3).reshape(h, bs * w)))  #             return value.transpose(1,0,2,3).reshape(h,bs*w) |> L
    if not _coconut_case_match_check_0:  # batch of rgbs  #         match (_,_,_,3): # batch of rgbs
        if (_coconut.isinstance(_coconut_case_match_to_0, _coconut.abc.Sequence)) and (_coconut.len(_coconut_case_match_to_0) == 4) and (_coconut_case_match_to_0[3] == 3):  # batch of rgbs  #         match (_,_,_,3): # batch of rgbs
            _coconut_case_match_check_0 = True  # batch of rgbs  #         match (_,_,_,3): # batch of rgbs
        if _coconut_case_match_check_0:  # batch of rgbs  #         match (_,_,_,3): # batch of rgbs
            return ((RGB)((batch_image_to_image)(value)))  #             return value |> batch_image_to_image |> RGB
    if not _coconut_case_match_check_0:  # batch of rgbas  #         match (_,_,_,4): # batch of rgbas
        if (_coconut.isinstance(_coconut_case_match_to_0, _coconut.abc.Sequence)) and (_coconut.len(_coconut_case_match_to_0) == 4) and (_coconut_case_match_to_0[3] == 4):  # batch of rgbas  #         match (_,_,_,4): # batch of rgbas
            _coconut_case_match_check_0 = True  # batch of rgbas  #         match (_,_,_,4): # batch of rgbas
        if _coconut_case_match_check_0:  # batch of rgbas  #         match (_,_,_,4): # batch of rgbas
            return ((RGBA)((batch_image_to_image)(value)))  #             return value |> batch_image_to_image |> RGBA
    if not _coconut_case_match_check_0:  # a gray image  #         match (_,_,1): # a gray image
        if (_coconut.isinstance(_coconut_case_match_to_0, _coconut.abc.Sequence)) and (_coconut.len(_coconut_case_match_to_0) == 3) and (_coconut_case_match_to_0[2] == 1):  # a gray image  #         match (_,_,1): # a gray image
            _coconut_case_match_check_0 = True  # a gray image  #         match (_,_,1): # a gray image
        if _coconut_case_match_check_0:  # a gray image  #         match (_,_,1): # a gray image
            return ((L)(value[:, :, 0]))  #             return value[:,:,0] |> L
    if not _coconut_case_match_check_0:  # an RGB image  #         match (_,_,3): # an RGB image
        if (_coconut.isinstance(_coconut_case_match_to_0, _coconut.abc.Sequence)) and (_coconut.len(_coconut_case_match_to_0) == 3) and (_coconut_case_match_to_0[2] == 3):  # an RGB image  #         match (_,_,3): # an RGB image
            _coconut_case_match_check_0 = True  # an RGB image  #         match (_,_,3): # an RGB image
        if _coconut_case_match_check_0:  # an RGB image  #         match (_,_,3): # an RGB image
            return ((RGB)(value))  #             return value |> RGB
    if not _coconut_case_match_check_0:  # an RGBA image  #         match (_,_,4): # an RGBA image
        if (_coconut.isinstance(_coconut_case_match_to_0, _coconut.abc.Sequence)) and (_coconut.len(_coconut_case_match_to_0) == 3) and (_coconut_case_match_to_0[2] == 4):  # an RGBA image  #         match (_,_,4): # an RGBA image
            _coconut_case_match_check_0 = True  # an RGBA image  #         match (_,_,4): # an RGBA image
        if _coconut_case_match_check_0:  # an RGBA image  #         match (_,_,4): # an RGBA image
            return ((RGBA)(value))  #             return value |> RGBA
    if not _coconut_case_match_check_0:  # batch of gray image  #         match (bs,h,w) if w>4: # batch of gray image
        _coconut_match_set_name_bs = _coconut_sentinel  # batch of gray image  #         match (bs,h,w) if w>4: # batch of gray image
        _coconut_match_set_name_h = _coconut_sentinel  # batch of gray image  #         match (bs,h,w) if w>4: # batch of gray image
        _coconut_match_set_name_w = _coconut_sentinel  # batch of gray image  #         match (bs,h,w) if w>4: # batch of gray image
        if (_coconut.isinstance(_coconut_case_match_to_0, _coconut.abc.Sequence)) and (_coconut.len(_coconut_case_match_to_0) == 3):  # batch of gray image  #         match (bs,h,w) if w>4: # batch of gray image
            _coconut_match_set_name_bs = _coconut_case_match_to_0[0]  # batch of gray image  #         match (bs,h,w) if w>4: # batch of gray image
            _coconut_match_set_name_h = _coconut_case_match_to_0[1]  # batch of gray image  #         match (bs,h,w) if w>4: # batch of gray image
            _coconut_match_set_name_w = _coconut_case_match_to_0[2]  # batch of gray image  #         match (bs,h,w) if w>4: # batch of gray image
            _coconut_case_match_check_0 = True  # batch of gray image  #         match (bs,h,w) if w>4: # batch of gray image
        if _coconut_case_match_check_0:  # batch of gray image  #         match (bs,h,w) if w>4: # batch of gray image
            if _coconut_match_set_name_bs is not _coconut_sentinel:  # batch of gray image  #         match (bs,h,w) if w>4: # batch of gray image
                bs = _coconut_case_match_to_0[0]  # batch of gray image  #         match (bs,h,w) if w>4: # batch of gray image
            if _coconut_match_set_name_h is not _coconut_sentinel:  # batch of gray image  #         match (bs,h,w) if w>4: # batch of gray image
                h = _coconut_case_match_to_0[1]  # batch of gray image  #         match (bs,h,w) if w>4: # batch of gray image
            if _coconut_match_set_name_w is not _coconut_sentinel:  # batch of gray image  #         match (bs,h,w) if w>4: # batch of gray image
                w = _coconut_case_match_to_0[2]  # batch of gray image  #         match (bs,h,w) if w>4: # batch of gray image
        if _coconut_case_match_check_0 and not (w > 4):  # batch of gray image  #         match (bs,h,w) if w>4: # batch of gray image
            _coconut_case_match_check_0 = False  # batch of gray image  #         match (bs,h,w) if w>4: # batch of gray image
        if _coconut_case_match_check_0:  # batch of gray image  #         match (bs,h,w) if w>4: # batch of gray image
            return ((L)(value.transpose(1, 0, 2).reshape(h, bs * w)))  #             return value.transpose(1,0,2).reshape(h,bs*w) |> L
    if not _coconut_case_match_check_0:  # gray image  #         match (_,_): # gray image
        if (_coconut.isinstance(_coconut_case_match_to_0, _coconut.abc.Sequence)) and (_coconut.len(_coconut_case_match_to_0) == 2):  # gray image  #         match (_,_): # gray image
            _coconut_case_match_check_0 = True  # gray image  #         match (_,_): # gray image
        if _coconut_case_match_check_0:  # gray image  #         match (_,_): # gray image
            return ((L)(value))  #             return value |> L
    if not _coconut_case_match_check_0:  #     else:
        return (None)  #         return None

def ary_stat_widget(ary):  # def ary_stat_widget(ary):
    return (widgets.VBox([widgets.Label(value=pformat(str(dict(shape=ary.shape, dtype=ary.dtype)))), widgets.HTML(pd.DataFrame(pd.Series(ary.ravel()).describe()).transpose()._repr_html_())]))  #     return widgets.VBox([


def ary_to_widget(ary):  # def ary_to_widget(ary):
    stat_widget = ary_stat_widget(ary)  #     stat_widget = ary_stat_widget(ary)
    viz_widget = (nn)(ary_to_image(ary), img_to_widget)  #     viz_widget = ary_to_image(ary) `nn` img_to_widget
    viz_widget = widgets.Label(value=str(ary)) if viz_widget is None else viz_widget  #     viz_widget ??= widgets.Label(value=str(ary))

    children = [viz_widget, (bqplot_hist)(ary), stat_widget]  #     children = [
    return (widgets.VBox(children))  #     return widgets.VBox(children)

def output_widget(value):  # def output_widget(value):
    out = widgets.Output()  #     out = widgets.Output()
    with out:  #     with out:
        pprint(repr(value))  #         pprint(repr(value))
    return (out)  #     return out

def infer_widget(value):  # def infer_widget(value):
    while True:  #     from data_tree._series import Series
        from data_tree._series import Series  #     from data_tree._series import Series
        from data_tree.coconut.convert import AutoImage  #     from data_tree.coconut.convert import AutoImage
        _coconut_case_match_to_1 = value  #     case value:
        _coconut_case_match_check_1 = False  #     case value:
        _coconut_case_match_check_1 = True  #     case value:
        if _coconut_case_match_check_1 and not (hasattr(value, "to_widget")):  #     case value:
            _coconut_case_match_check_1 = False  #     case value:
        if _coconut_case_match_check_1:  #     case value:
            return (value.to_widget())  #             return value.to_widget()
        if not _coconut_case_match_check_1:  #         match _ is AutoImage:
            if _coconut.isinstance(_coconut_case_match_to_1, AutoImage):  #         match _ is AutoImage:
                _coconut_case_match_check_1 = True  #         match _ is AutoImage:
            if _coconut_case_match_check_1:  #         match _ is AutoImage:
                return (value.to_widget())  #             return value.to_widget()
        if not _coconut_case_match_check_1:  #         match _ is Series:
            if _coconut.isinstance(_coconut_case_match_to_1, Series):  #         match _ is Series:
                _coconut_case_match_check_1 = True  #         match _ is Series:
            if _coconut_case_match_check_1:  #         match _ is Series:
                return (value.widget())  #             return value.widget()
        if not _coconut_case_match_check_1:  #         match _ is PIL.Image.Image:
            if _coconut.isinstance(_coconut_case_match_to_1, PIL.Image.Image):  #         match _ is PIL.Image.Image:
                _coconut_case_match_check_1 = True  #         match _ is PIL.Image.Image:
            if _coconut_case_match_check_1:  #         match _ is PIL.Image.Image:
                return ((img_to_widget)(value))  #             return value |> img_to_widget
        if not _coconut_case_match_check_1:  #         match _ is TensorType:
            if _coconut.isinstance(_coconut_case_match_to_1, TensorType):  #         match _ is TensorType:
                _coconut_case_match_check_1 = True  #         match _ is TensorType:
            if _coconut_case_match_check_1:  #         match _ is TensorType:
                try:  #         match _ is TensorType:
                    _coconut_tre_check_0 = infer_widget is _coconut_recursive_func_6  #         match _ is TensorType:
                except _coconut.NameError:  #         match _ is TensorType:
                    _coconut_tre_check_0 = False  #         match _ is TensorType:
                if _coconut_tre_check_0:  #         match _ is TensorType:
                    value = value.detach().cpu().numpy()  #         match _ is TensorType:
                    continue  #         match _ is TensorType:
                else:  #         match _ is TensorType:
                    return infer_widget(value.detach().cpu().numpy())  #         match _ is TensorType:

        if not _coconut_case_match_check_1:  #         match _ is np.ndarray:
            if _coconut.isinstance(_coconut_case_match_to_1, np.ndarray):  #         match _ is np.ndarray:
                _coconut_case_match_check_1 = True  #         match _ is np.ndarray:
            if _coconut_case_match_check_1:  #         match _ is np.ndarray:
                return (ary_to_widget(value))  #             return ary_to_widget(value)
        if not _coconut_case_match_check_1:  #         match _ is tuple if value `hasattr` "_asdict":
            if _coconut.isinstance(_coconut_case_match_to_1, tuple):  #         match _ is tuple if value `hasattr` "_asdict":
                _coconut_case_match_check_1 = True  #         match _ is tuple if value `hasattr` "_asdict":
            if _coconut_case_match_check_1 and not ((hasattr)(value, "_asdict")):  #         match _ is tuple if value `hasattr` "_asdict":
                _coconut_case_match_check_1 = False  #         match _ is tuple if value `hasattr` "_asdict":
            if _coconut_case_match_check_1:  #         match _ is tuple if value `hasattr` "_asdict":
                return ((output_widget)(value))  #             return value |> output_widget
        if not _coconut_case_match_check_1:  #         match _ is dict:
            if _coconut.isinstance(_coconut_case_match_to_1, dict):  #         match _ is dict:
                _coconut_case_match_check_1 = True  #         match _ is dict:
            if _coconut_case_match_check_1:  #         match _ is dict:
                return (widgets.VBox([widgets.VBox([widgets.Text(k), infer_widget(v)]) for k, v in value.items()]))  #             return widgets.VBox([
        if not _coconut_case_match_check_1:  #                 widgets.VBox([widgets.Text(k),infer_widget(v)]) for k,v in value.items()
            if _coconut.isinstance(_coconut_case_match_to_1, (tuple, list)):  #                 widgets.VBox([widgets.Text(k),infer_widget(v)]) for k,v in value.items()
                _coconut_case_match_check_1 = True  #                 widgets.VBox([widgets.Text(k),infer_widget(v)]) for k,v in value.items()
            if _coconut_case_match_check_1:  #                 widgets.VBox([widgets.Text(k),infer_widget(v)]) for k,v in value.items()
                items = [infer_widget(item) for item in value]  #             items = [infer_widget(item) for item in value]
                return (widgets.VBox([widgets.GridBox(items, layout=widgets.Layout(grid_template_columns="auto auto auto", border="solid 2px")), widgets.Label(value="displaying tuple with {_coconut_format_0} elements".format(_coconut_format_0=(len(value))))]))  #             return widgets.VBox([
        if not _coconut_case_match_check_1:  #     else:
            return ((output_widget)(value))  #         return value |> output_widget


        return None  # 
_coconut_recursive_func_6 = infer_widget  #



import base64
from io import BytesIO

import PIL
from omni_converter.solver.rules import AutoRuleBook
from toolz.functoolz import memoize



def make_rule(src, dst):
    def _inner(converter):
        def rule_impl(s):
            if s == src:
                return [(converter, dst)]

        return rule_impl

    return _inner




@make_rule("image,RGB,RGB", "base64_png")
def pil_img_to_base64(img: PIL.Image.Image):
    buf = BytesIO()
    img.save(buf, format="png")
    img_str = base64.b64encode(buf.getvalue()).decode("ascii")
    return img_str


@make_rule("base64_png", "html")
def base64_png_to_html(base64_str):
    return f"""<img src="data:image/png;base64,{base64_str}"/>"""


@make_rule("dataframe", "html")
def custom_dataframe_viz(df):
    def _formatter(item):
        if hasattr(item, "_repr_html_"):
            return item._repr_html_()
        else:
            return repr(item)

    formatters = {k: _formatter for k in df.columns}
    return df.to_html(escape=False, formatters=formatters)


custom_rule_book = AutoRuleBook(id="custom_rule_book").add_rules(
    pil_img_to_base64,
    base64_png_to_html,
    custom_dataframe_viz
).add_alias("html", "_repr_html_").add_alias("base64","base64_png").set_id("custom_rule_book")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# type: ignore

# Compiled with Coconut version 1.6.0 [Vocational Guidance Counsellor]

"""Built-in Coconut utilities."""

# Coconut Header: -------------------------------------------------------------

from __future__ import generator_stop
import sys as _coconut_sys
from builtins import chr, filter, hex, input, int, map, object, oct, open, print, range, str, zip, filter, reversed, enumerate
py_chr, py_hex, py_input, py_int, py_map, py_object, py_oct, py_open, py_print, py_range, py_str, py_zip, py_filter, py_reversed, py_enumerate, py_repr = chr, hex, input, int, map, object, oct, open, print, range, str, zip, filter, reversed, enumerate, repr
_coconut_py_str = str
exec("_coconut_exec = exec")
if _coconut_sys.version_info < (3, 7):
    def _coconut_default_breakpointhook(*args, **kwargs):
        hookname = _coconut.os.getenv("PYTHONBREAKPOINT")
        if hookname != "0":
            if not hookname:
                hookname = "pdb.set_trace"
            modname, dot, funcname = hookname.rpartition(".")
            if not dot:
                modname = "builtins" if _coconut_sys.version_info >= (3,) else "__builtin__"
            if _coconut_sys.version_info >= (2, 7):
                import importlib
                module = importlib.import_module(modname)
            else:
                import imp
                module = imp.load_module(modname, *imp.find_module(modname))
            hook = _coconut.getattr(module, funcname)
            return hook(*args, **kwargs)
    if not hasattr(_coconut_sys, "__breakpointhook__"):
        _coconut_sys.__breakpointhook__ = _coconut_default_breakpointhook
    def breakpoint(*args, **kwargs):
        return _coconut.getattr(_coconut_sys, "breakpointhook", _coconut_default_breakpointhook)(*args, **kwargs)
else:
    py_breakpoint = breakpoint
class _coconut:
    import collections, copy, functools, types, itertools, operator, threading, os, warnings, contextlib, traceback, weakref, multiprocessing
    from multiprocessing import dummy as multiprocessing_dummy
    import asyncio
    import pickle
    OrderedDict = collections.OrderedDict
    import collections.abc as abc
    if _coconut_sys.version_info < (3, 6):
        class typing:
            @staticmethod
            def NamedTuple(name, fields):
                return _coconut.collections.namedtuple(name, [x for x, t in fields])
    else:
        import typing
    zip_longest = itertools.zip_longest
    Ellipsis, NotImplemented, NotImplementedError, Exception, AttributeError, ImportError, IndexError, KeyError, NameError, TypeError, ValueError, StopIteration, RuntimeError, any, bytes, classmethod, dict, enumerate, filter, float, frozenset, getattr, hasattr, hash, id, int, isinstance, issubclass, iter, len, list, locals, map, min, max, next, object, property, range, reversed, set, slice, str, sum, super, tuple, type, vars, zip, repr, print = Ellipsis, NotImplemented, NotImplementedError, Exception, AttributeError, ImportError, IndexError, KeyError, NameError, TypeError, ValueError, StopIteration, RuntimeError, any, bytes, classmethod, dict, enumerate, filter, float, frozenset, getattr, hasattr, hash, id, int, isinstance, issubclass, iter, len, list, locals, map, min, max, next, object, property, range, reversed, set, slice, str, sum, super, tuple, type, vars, zip, repr, print
class _coconut_sentinel: pass
class _coconut_base_hashable:
    __slots__ = ()
    def __reduce_ex__(self, _):
        return self.__reduce__()
    def __eq__(self, other):
        return self.__class__ is other.__class__ and self.__reduce__() == other.__reduce__()
    def __hash__(self):
        return _coconut.hash(self.__reduce__())
class MatchError(_coconut_base_hashable, Exception):
    """Pattern-matching error. Has attributes .pattern, .value, and .message."""
    __slots__ = ("pattern", "value", "_message")
    max_val_repr_len = 500
    def __init__(self, pattern=None, value=None):
        self.pattern = pattern
        self.value = value
        self._message = None
    @property
    def message(self):
        if self._message is None:
            value_repr = _coconut.repr(self.value)
            self._message = "pattern-matching failed for %s in %s" % (_coconut.repr(self.pattern), value_repr if _coconut.len(value_repr) <= self.max_val_repr_len else value_repr[:self.max_val_repr_len] + "...")
            Exception.__init__(self, self._message)
        return self._message
    def __repr__(self):
        self.message
        return Exception.__repr__(self)
    def __str__(self):
        self.message
        return Exception.__str__(self)
    def __unicode__(self):
        self.message
        return Exception.__unicode__(self)
    def __reduce__(self):
        return (self.__class__, (self.pattern, self.value))
class _coconut_tail_call:
    __slots__ = ("func", "args", "kwargs")
    def __init__(self, func, *args, **kwargs):
        self.func = func
        self.args = args
        self.kwargs = kwargs
_coconut_tco_func_dict = {}
def _coconut_tco(func):
    @_coconut.functools.wraps(func)
    def tail_call_optimized_func(*args, **kwargs):
        call_func = func
        while True:
            if _coconut.isinstance(call_func, _coconut_base_pattern_func):
                call_func = call_func._coconut_tco_func
            elif _coconut.isinstance(call_func, _coconut.types.MethodType):
                wkref = _coconut_tco_func_dict.get(_coconut.id(call_func.__func__))
                wkref_func = None if wkref is None else wkref()
                if wkref_func is call_func.__func__:
                    if call_func.__self__ is None:
                        call_func = call_func._coconut_tco_func
                    else:
                        call_func = _coconut.functools.partial(call_func._coconut_tco_func, call_func.__self__)
            else:
                wkref = _coconut_tco_func_dict.get(_coconut.id(call_func))
                wkref_func = None if wkref is None else wkref()
                if wkref_func is call_func:
                    call_func = call_func._coconut_tco_func
            result = call_func(*args, **kwargs)  # pass --no-tco to clean up your traceback
            if not isinstance(result, _coconut_tail_call):
                return result
            call_func, args, kwargs = result.func, result.args, result.kwargs
    tail_call_optimized_func._coconut_tco_func = func
    tail_call_optimized_func.__module__ = _coconut.getattr(func, "__module__", None)
    tail_call_optimized_func.__name__ = _coconut.getattr(func, "__name__", None)
    tail_call_optimized_func.__qualname__ = _coconut.getattr(func, "__qualname__", None)
    _coconut_tco_func_dict[_coconut.id(tail_call_optimized_func)] = _coconut.weakref.ref(tail_call_optimized_func)
    return tail_call_optimized_func
def _coconut_igetitem(iterable, index):
    obj_igetitem = _coconut.getattr(iterable, "__iter_getitem__", None)
    if obj_igetitem is None:
        obj_igetitem = _coconut.getattr(iterable, "__getitem__", None)
    if obj_igetitem is not None:
        try:
            result = obj_igetitem(index)
        except _coconut.NotImplementedError:
            pass
        else:
            if result is not _coconut.NotImplemented:
                return result
    if not _coconut.isinstance(index, _coconut.slice):
        if index < 0:
            return _coconut.collections.deque(iterable, maxlen=-index)[0]
        try:
            return _coconut.next(_coconut.itertools.islice(iterable, index, index + 1))
        except _coconut.StopIteration:
            raise _coconut.IndexError("$[] index out of range")
    if index.start is not None and index.start < 0 and (index.stop is None or index.stop < 0) and index.step is None:
        queue = _coconut.collections.deque(iterable, maxlen=-index.start)
        if index.stop is not None:
            queue = _coconut.list(queue)[:index.stop - index.start]
        return queue
    if (index.start is None or index.start == 0) and index.stop is None and index.step is not None and index.step == -1:
        obj_reversed = _coconut.getattr(iterable, "__reversed__", None)
        if obj_reversed is not None:
            try:
                result = obj_reversed()
            except _coconut.NotImplementedError:
                pass
            else:
                if result is not _coconut.NotImplemented:
                    return result
    if (index.start is not None and index.start < 0) or (index.stop is not None and index.stop < 0) or (index.step is not None and index.step < 0):
        return _coconut.list(iterable)[index]
    return _coconut.itertools.islice(iterable, index.start, index.stop, index.step)
class _coconut_base_compose(_coconut_base_hashable):
    __slots__ = ("func", "funcstars")
    def __init__(self, func, *funcstars):
        self.func = func
        self.funcstars = []
        for f, stars in funcstars:
            if _coconut.isinstance(f, _coconut_base_compose):
                self.funcstars.append((f.func, stars))
                self.funcstars += f.funcstars
            else:
                self.funcstars.append((f, stars))
        self.funcstars = _coconut.tuple(self.funcstars)
    def __call__(self, *args, **kwargs):
        arg = self.func(*args, **kwargs)
        for f, stars in self.funcstars:
            if stars == 0:
                arg = f(arg)
            elif stars == 1:
                arg = f(*arg)
            elif stars == 2:
                arg = f(**arg)
            else:
                raise _coconut.ValueError("invalid arguments to " + _coconut.repr(self))
        return arg
    def __repr__(self):
        return _coconut.repr(self.func) + " " + " ".join(("..*> " if star == 1 else "..**>" if star == 2 else "..> ") + _coconut.repr(f) for f, star in self.funcstars)
    def __reduce__(self):
        return (self.__class__, (self.func,) + self.funcstars)
    def __get__(self, obj, objtype=None):
        if obj is None:
            return self
        return _coconut.types.MethodType(self, obj)
def _coconut_forward_compose(func, *funcs): return _coconut_base_compose(func, *((f, 0) for f in funcs))
def _coconut_back_compose(*funcs): return _coconut_forward_compose(*_coconut.reversed(funcs))
def _coconut_forward_star_compose(func, *funcs): return _coconut_base_compose(func, *((f, 1) for f in funcs))
def _coconut_back_star_compose(*funcs): return _coconut_forward_star_compose(*_coconut.reversed(funcs))
def _coconut_forward_dubstar_compose(func, *funcs): return _coconut_base_compose(func, *((f, 2) for f in funcs))
def _coconut_back_dubstar_compose(*funcs): return _coconut_forward_dubstar_compose(*_coconut.reversed(funcs))
def _coconut_pipe(x, f): return f(x)
def _coconut_star_pipe(xs, f): return f(*xs)
def _coconut_dubstar_pipe(kws, f): return f(**kws)
def _coconut_back_pipe(f, x): return f(x)
def _coconut_back_star_pipe(f, xs): return f(*xs)
def _coconut_back_dubstar_pipe(f, kws): return f(**kws)
def _coconut_none_pipe(x, f): return None if x is None else f(x)
def _coconut_none_star_pipe(xs, f): return None if xs is None else f(*xs)
def _coconut_none_dubstar_pipe(kws, f): return None if kws is None else f(**kws)
def _coconut_assert(cond, msg=None): assert cond, msg if msg is not None else "(assert) got falsey value " + _coconut.repr(cond)
def _coconut_bool_and(a, b): return a and b
def _coconut_bool_or(a, b): return a or b
def _coconut_none_coalesce(a, b): return a if a is not None else b
def _coconut_minus(a, *rest):
    if not rest:
        return -a
    for b in rest:
        a = a - b
    return a
@_coconut.functools.wraps(_coconut.itertools.tee)
def tee(iterable, n=2):
    if n >= 0 and _coconut.isinstance(iterable, (_coconut.tuple, _coconut.frozenset)):
        return (iterable,) * n
    if n > 0 and (_coconut.isinstance(iterable, _coconut.abc.Sequence) or _coconut.getattr(iterable, "__copy__", None) is not None):
        return (iterable,) + _coconut.tuple(_coconut.copy.copy(iterable) for _ in _coconut.range(n - 1))
    return _coconut.itertools.tee(iterable, n)
class reiterable(_coconut_base_hashable):
    """Allows an iterator to be iterated over multiple times."""
    __slots__ = ("lock", "iter")
    def __new__(cls, iterable):
        if _coconut.isinstance(iterable, _coconut_reiterable):
            return iterable
        self = _coconut.object.__new__(cls)
        self.lock = _coconut.threading.Lock()
        self.iter = iterable
        return self
    def get_new_iter(self):
        with self.lock:
            self.iter, new_iter = _coconut_tee(self.iter)
        return new_iter
    def __iter__(self):
        return _coconut.iter(self.get_new_iter())
    def __getitem__(self, index):
        return _coconut_igetitem(self.get_new_iter(), index)
    def __reversed__(self):
        return _coconut_reversed(self.get_new_iter())
    def __len__(self):
        return _coconut.len(self.iter)
    def __repr__(self):
        return "reiterable(%r)" % (self.iter,)
    def __reduce__(self):
        return (self.__class__, (self.iter,))
    def __copy__(self):
        return self.__class__(self.get_new_iter())
    def __fmap__(self, func):
        return _coconut_map(func, self)
class scan(_coconut_base_hashable):
    """Reduce func over iterable, yielding intermediate results,
    optionally starting from initializer."""
    __slots__ = ("func", "iter", "initializer")
    def __init__(self, function, iterable, initializer=_coconut_sentinel):
        self.func = function
        self.iter = iterable
        self.initializer = initializer
    def __iter__(self):
        acc = self.initializer
        if acc is not _coconut_sentinel:
            yield acc
        for item in self.iter:
            if acc is _coconut_sentinel:
                acc = item
            else:
                acc = self.func(acc, item)
            yield acc
    def __len__(self):
        return _coconut.len(self.iter)
    def __repr__(self):
        return "scan(%r, %r)" % (self.func, self.iter) if self.initializer is _coconut_sentinel else "scan(%r, %r, %r)" % (self.func, self.iter, self.initializer)
    def __reduce__(self):
        return (self.__class__, (self.func, self.iter, self.initializer))
    def __fmap__(self, func):
        return _coconut_map(func, self)
class reversed(_coconut_base_hashable):
    __slots__ = ("iter",)
    if hasattr(_coconut.map, "__doc__"):
        __doc__ = _coconut.reversed.__doc__
    def __new__(cls, iterable):
        if _coconut.isinstance(iterable, _coconut.range):
            return iterable[::-1]
        if not _coconut.hasattr(iterable, "__reversed__") or _coconut.isinstance(iterable, (_coconut.list, _coconut.tuple)):
            return _coconut.object.__new__(cls)
        return _coconut.reversed(iterable)
    def __init__(self, iterable):
        self.iter = iterable
    def __iter__(self):
        return _coconut.iter(_coconut.reversed(self.iter))
    def __getitem__(self, index):
        if _coconut.isinstance(index, _coconut.slice):
            return _coconut_igetitem(self.iter, _coconut.slice(-(index.start + 1) if index.start is not None else None, -(index.stop + 1) if index.stop else None, -(index.step if index.step is not None else 1)))
        return _coconut_igetitem(self.iter, -(index + 1))
    def __reversed__(self):
        return self.iter
    def __len__(self):
        return _coconut.len(self.iter)
    def __repr__(self):
        return "reversed(%r)" % (self.iter,)
    def __reduce__(self):
        return (self.__class__, (self.iter,))
    def __contains__(self, elem):
        return elem in self.iter
    def count(self, elem):
        """Count the number of times elem appears in the reversed iterable."""
        return self.iter.count(elem)
    def index(self, elem):
        """Find the index of elem in the reversed iterable."""
        return _coconut.len(self.iter) - self.iter.index(elem) - 1
    def __fmap__(self, func):
        return self.__class__(_coconut_map(func, self.iter))
class flatten(_coconut_base_hashable):
    """Flatten an iterable of iterables into a single iterable."""
    __slots__ = ("iter",)
    def __init__(self, iterable):
        self.iter = iterable
    def __iter__(self):
        return _coconut.itertools.chain.from_iterable(self.iter)
    def __reversed__(self):
        return self.__class__(_coconut_reversed(_coconut_map(_coconut_reversed, self.iter)))
    def __repr__(self):
        return "flatten(%r)" % (self.iter,)
    def __reduce__(self):
        return (self.__class__, (self.iter,))
    def __contains__(self, elem):
        self.iter, new_iter = _coconut_tee(self.iter)
        return _coconut.any(elem in it for it in new_iter)
    def count(self, elem):
        """Count the number of times elem appears in the flattened iterable."""
        self.iter, new_iter = _coconut_tee(self.iter)
        return _coconut.sum(it.count(elem) for it in new_iter)
    def index(self, elem):
        self.iter, new_iter = _coconut_tee(self.iter)
        ind = 0
        for it in new_iter:
            try:
                return ind + it.index(elem)
            except _coconut.ValueError:
                ind += _coconut.len(it)
        raise ValueError("%r not in %r" % (elem, self))
    def __fmap__(self, func):
        return self.__class__(_coconut_map(_coconut.functools.partial(_coconut_map, func), self.iter))
class map(_coconut_base_hashable, _coconut.map):
    __slots__ = ("func", "iters")
    if hasattr(_coconut.map, "__doc__"):
        __doc__ = _coconut.map.__doc__
    def __new__(cls, function, *iterables):
        new_map = _coconut.map.__new__(cls, function, *iterables)
        new_map.func = function
        new_map.iters = iterables
        return new_map
    def __getitem__(self, index):
        if _coconut.isinstance(index, _coconut.slice):
            return self.__class__(self.func, *(_coconut_igetitem(i, index) for i in self.iters))
        return self.func(*(_coconut_igetitem(i, index) for i in self.iters))
    def __reversed__(self):
        return self.__class__(self.func, *(_coconut_reversed(i) for i in self.iters))
    def __len__(self):
        return _coconut.min(_coconut.len(i) for i in self.iters)
    def __repr__(self):
        return "map(%r, %s)" % (self.func, ", ".join((_coconut.repr(i) for i in self.iters)))
    def __reduce__(self):
        return (self.__class__, (self.func,) + self.iters)
    def __iter__(self):
        return _coconut.iter(_coconut.map(self.func, *self.iters))
    def __fmap__(self, func):
        return self.__class__(_coconut_forward_compose(self.func, func), *self.iters)
class _coconut_parallel_concurrent_map_func_wrapper(_coconut_base_hashable):
    __slots__ = ("map_cls", "func", "star")
    def __init__(self, map_cls, func, star):
        self.map_cls = map_cls
        self.func = func
        self.star = star
    def __reduce__(self):
        return (self.__class__, (self.map_cls, self.func, self.star))
    def __call__(self, *args, **kwargs):
        self.map_cls.get_pool_stack().append(None)
        try:
            if self.star:
                assert _coconut.len(args) == 1, "internal parallel/concurrent map error"
                return self.func(*args[0], **kwargs)
            else:
                return self.func(*args, **kwargs)
        except:
            _coconut.print(self.map_cls.__name__ + " error:")
            _coconut.traceback.print_exc()
            raise
        finally:
            assert self.map_cls.get_pool_stack().pop() is None, "internal parallel/concurrent map error"
class _coconut_base_parallel_concurrent_map(map):
    __slots__ = ("result", "chunksize")
    @classmethod
    def get_pool_stack(cls):
        return cls.threadlocal_ns.__dict__.setdefault("pool_stack", [None])
    def __new__(cls, function, *iterables, **kwargs):
        self = _coconut_map.__new__(cls, function, *iterables)
        self.result = None
        self.chunksize = kwargs.pop("chunksize", 1)
        if kwargs:
            raise _coconut.TypeError(cls.__name__ + "() got unexpected keyword arguments " + _coconut.repr(kwargs))
        if cls.get_pool_stack()[-1] is not None:
            return self.get_list()
        return self
    @classmethod
    @_coconut.contextlib.contextmanager
    def multiple_sequential_calls(cls, max_workers=None):
        """Context manager that causes nested calls to use the same pool."""
        if cls.get_pool_stack()[-1] is None:
            cls.get_pool_stack()[-1] = cls.make_pool(max_workers)
            try:
                yield
            finally:
                cls.get_pool_stack()[-1].terminate()
                cls.get_pool_stack()[-1] = None
        else:
            yield
    def get_list(self):
        if self.result is None:
            with self.multiple_sequential_calls():
                if _coconut.len(self.iters) == 1:
                    self.result = _coconut.list(self.get_pool_stack()[-1].imap(_coconut_parallel_concurrent_map_func_wrapper(self.__class__, self.func, False), self.iters[0], self.chunksize))
                else:
                    self.result = _coconut.list(self.get_pool_stack()[-1].imap(_coconut_parallel_concurrent_map_func_wrapper(self.__class__, self.func, True), _coconut.zip(*self.iters), self.chunksize))
        return self.result
    def __iter__(self):
        return _coconut.iter(self.get_list())
class parallel_map(_coconut_base_parallel_concurrent_map):
    """
    Multi-process implementation of map. Requires arguments to be pickleable.
    For multiple sequential calls, use:
        with parallel_map.multiple_sequential_calls():
            ...
    """
    __slots__ = ()
    threadlocal_ns = _coconut.threading.local()
    @staticmethod
    def make_pool(max_workers=None):
        return _coconut.multiprocessing.Pool(max_workers)
    def __repr__(self):
        return "parallel_" + _coconut_map.__repr__(self)
class concurrent_map(_coconut_base_parallel_concurrent_map):
    """
    Multi-thread implementation of map. For multiple sequential calls, use:
        with concurrent_map.multiple_sequential_calls():
            ...
    """
    __slots__ = ()
    threadlocal_ns = _coconut.threading.local()
    @staticmethod
    def make_pool(max_workers=None):
        return _coconut.multiprocessing_dummy.Pool(_coconut.multiprocessing.cpu_count() * 5 if max_workers is None else max_workers)
    def __repr__(self):
        return "concurrent_" + _coconut_map.__repr__(self)
class filter(_coconut_base_hashable, _coconut.filter):
    __slots__ = ("func", "iter")
    if hasattr(_coconut.filter, "__doc__"):
        __doc__ = _coconut.filter.__doc__
    def __new__(cls, function, iterable):
        new_filter = _coconut.filter.__new__(cls, function, iterable)
        new_filter.func = function
        new_filter.iter = iterable
        return new_filter
    def __reversed__(self):
        return self.__class__(self.func, _coconut_reversed(self.iter))
    def __repr__(self):
        return "filter(%r, %r)" % (self.func, self.iter)
    def __reduce__(self):
        return (self.__class__, (self.func, self.iter))
    def __iter__(self):
        return _coconut.iter(_coconut.filter(self.func, self.iter))
    def __fmap__(self, func):
        return _coconut_map(func, self)
class zip(_coconut_base_hashable, _coconut.zip):
    __slots__ = ("iters", "strict")
    if hasattr(_coconut.zip, "__doc__"):
        __doc__ = _coconut.zip.__doc__
    def __new__(cls, *iterables, **kwargs):
        new_zip = _coconut.zip.__new__(cls, *iterables)
        new_zip.iters = iterables
        new_zip.strict = kwargs.pop("strict", False)
        if kwargs:
            raise _coconut.TypeError("zip() got unexpected keyword arguments " + _coconut.repr(kwargs))
        return new_zip
    def __getitem__(self, index):
        if _coconut.isinstance(index, _coconut.slice):
            return self.__class__(*(_coconut_igetitem(i, index) for i in self.iters), strict=self.strict)
        return _coconut.tuple(_coconut_igetitem(i, index) for i in self.iters)
    def __reversed__(self):
        return self.__class__(*(_coconut_reversed(i) for i in self.iters), strict=self.strict)
    def __len__(self):
        return _coconut.min(_coconut.len(i) for i in self.iters)
    def __repr__(self):
        return "zip(%s%s)" % (", ".join((_coconut.repr(i) for i in self.iters)), ", strict=True" if self.strict else "")
    def __reduce__(self):
        return (self.__class__, self.iters, self.strict)
    def __setstate__(self, strict):
        self.strict = strict
    def __iter__(self):
        for items in _coconut.iter(_coconut.zip_longest(*self.iters, fillvalue=_coconut_sentinel) if self.strict else _coconut.zip(*self.iters)):
            if self.strict and _coconut.any(x is _coconut_sentinel for x in items):
                raise _coconut.ValueError("zip(..., strict=True) arguments have mismatched lengths")
            yield items
    def __fmap__(self, func):
        return _coconut_map(func, self)
class zip_longest(zip):
    __slots__ = ("fillvalue",)
    if hasattr(_coconut.zip_longest, "__doc__"):
        __doc__ = (_coconut.zip_longest).__doc__
    def __new__(cls, *iterables, **kwargs):
        self = _coconut_zip.__new__(cls, *iterables, strict=False)
        self.fillvalue = kwargs.pop("fillvalue", None)
        if kwargs:
            raise _coconut.TypeError("zip_longest() got unexpected keyword arguments " + _coconut.repr(kwargs))
        return self
    def __getitem__(self, index):
        if _coconut.isinstance(index, _coconut.slice):
            new_ind = _coconut.slice(index.start + self.__len__() if index.start is not None and index.start < 0 else index.start, index.stop + self.__len__() if index.stop is not None and index.stop < 0 else index.stop, index.step)
            return self.__class__(*(_coconut_igetitem(i, new_ind) for i in self.iters))
        if index < 0:
            index += self.__len__()
        result = []
        got_non_default = False
        for it in self.iters:
            try:
                result.append(_coconut_igetitem(it, index))
            except _coconut.IndexError:
                result.append(self.fillvalue)
            else:
                got_non_default = True
        if not got_non_default:
            raise _coconut.IndexError("zip_longest index out of range")
        return _coconut.tuple(result)
    def __len__(self):
        return _coconut.max(_coconut.len(i) for i in self.iters)
    def __repr__(self):
        return "zip_longest(%s, fillvalue=%s)" % (", ".join((_coconut.repr(i) for i in self.iters)), self.fillvalue)
    def __reduce__(self):
        return (self.__class__, self.iters, self.fillvalue)
    def __setstate__(self, fillvalue):
        self.fillvalue = fillvalue
    def __iter__(self):
        return _coconut.iter(_coconut.zip_longest(*self.iters, fillvalue=self.fillvalue))
class enumerate(_coconut_base_hashable, _coconut.enumerate):
    __slots__ = ("iter", "start")
    if hasattr(_coconut.enumerate, "__doc__"):
        __doc__ = _coconut.enumerate.__doc__
    def __new__(cls, iterable, start=0):
        new_enumerate = _coconut.enumerate.__new__(cls, iterable, start)
        new_enumerate.iter = iterable
        new_enumerate.start = start
        return new_enumerate
    def __getitem__(self, index):
        if _coconut.isinstance(index, _coconut.slice):
            return self.__class__(_coconut_igetitem(self.iter, index), self.start + (0 if index.start is None else index.start if index.start >= 0 else _coconut.len(self.iter) + index.start))
        return (self.start + index, _coconut_igetitem(self.iter, index))
    def __len__(self):
        return _coconut.len(self.iter)
    def __repr__(self):
        return "enumerate(%r, %r)" % (self.iter, self.start)
    def __reduce__(self):
        return (self.__class__, (self.iter, self.start))
    def __iter__(self):
        return _coconut.iter(_coconut.enumerate(self.iter, self.start))
    def __fmap__(self, func):
        return _coconut_map(func, self)
class count(_coconut_base_hashable):
    """count(start, step) returns an infinite iterator starting at start and increasing by step.
    If step is set to 0, count will infinitely repeat its first argument."""
    __slots__ = ("start", "step")
    def __init__(self, start=0, step=1):
        self.start = start
        self.step = step
    def __iter__(self):
        while True:
            yield self.start
            if self.step:
                self.start += self.step
    def __contains__(self, elem):
        if not self.step:
            return elem == self.start
        if elem < self.start:
            return False
        return (elem - self.start) % self.step == 0
    def __getitem__(self, index):
        if _coconut.isinstance(index, _coconut.slice) and (index.start is None or index.start >= 0) and (index.stop is None or index.stop >= 0):
            new_start, new_step = self.start, self.step
            if self.step and index.start is not None:
                new_start += self.step * index.start
            if self.step and index.step is not None:
                new_step *= index.step
            if index.stop is None:
                return self.__class__(new_start, new_step)
            if self.step and _coconut.isinstance(self.start, _coconut.int) and _coconut.isinstance(self.step, _coconut.int):
                return _coconut.range(new_start, self.start + self.step * index.stop, new_step)
            return _coconut_map(self.__getitem__, _coconut.range(index.start if index.start is not None else 0, index.stop, index.step if index.step is not None else 1))
        if index < 0:
            raise _coconut.IndexError("count indices must be positive")
        return self.start + self.step * index if self.step else self.start
    def count(self, elem):
        """Count the number of times elem appears in the count."""
        if not self.step:
            return _coconut.float("inf") if elem == self.start else 0
        return int(elem in self)
    def index(self, elem):
        """Find the index of elem in the count."""
        if elem not in self:
            raise _coconut.ValueError(_coconut.repr(elem) + " not in " + _coconut.repr(self))
        return (elem - self.start) // self.step if self.step else 0
    def __reversed__(self):
        if not self.step:
            return self
        raise _coconut.TypeError(_coconut.repr(self) + " object is not reversible")
    def __repr__(self):
        return "count(%r, %r)" % (self.start, self.step)
    def __reduce__(self):
        return (self.__class__, (self.start, self.step))
    def __copy__(self):
        return self.__class__(self.start, self.step)
    def __fmap__(self, func):
        return _coconut_map(func, self)
class groupsof(_coconut_base_hashable):
    """groupsof(n, iterable) splits iterable into groups of size n.
    If the length of the iterable is not divisible by n, the last group may be of size < n."""
    __slots__ = ("group_size", "iter")
    def __init__(self, n, iterable):
        self.iter = iterable
        try:
            self.group_size = _coconut.int(n)
        except _coconut.ValueError:
            raise _coconut.TypeError("group size must be an int; not %r" % (n,))
        if self.group_size <= 0:
            raise _coconut.ValueError("group size must be > 0; not %r" % (self.group_size,))
    def __iter__(self):
        iterator = _coconut.iter(self.iter)
        loop = True
        while loop:
            group = []
            for _ in _coconut.range(self.group_size):
                try:
                    group.append(_coconut.next(iterator))
                except _coconut.StopIteration:
                    loop = False
                    break
            if group:
                yield _coconut.tuple(group)
    def __len__(self):
        return _coconut.len(self.iter)
    def __repr__(self):
        return "groupsof(%r)" % (self.iter,)
    def __reduce__(self):
        return (self.__class__, (self.group_size, self.iter))
    def __fmap__(self, func):
        return _coconut_map(func, self)
class recursive_iterator(_coconut_base_hashable):
    """Decorator that optimizes a function for iterator recursion."""
    __slots__ = ("func", "tee_store", "backup_tee_store")
    def __init__(self, func):
        self.func = func
        self.tee_store = {}
        self.backup_tee_store = []
    def __call__(self, *args, **kwargs):
        key = (args, _coconut.frozenset(kwargs.items()))
        use_backup = False
        try:
            _coconut.hash(key)
        except _coconut.Exception:
            try:
                key = _coconut.pickle.dumps(key, -1)
            except _coconut.Exception:
                use_backup = True
        if use_backup:
            for i, (k, v) in _coconut.enumerate(self.backup_tee_store):
                if k == key:
                    to_tee, store_pos = v, i
                    break
            else:
                to_tee = self.func(*args, **kwargs)
                store_pos = None
            to_store, to_return = _coconut_tee(to_tee)
            if store_pos is None:
                self.backup_tee_store.append([key, to_store])
            else:
                self.backup_tee_store[store_pos][1] = to_store
        else:
            it = self.tee_store.get(key)
            if it is None:
                it = self.func(*args, **kwargs)
            self.tee_store[key], to_return = _coconut_tee(it)
        return to_return
    def __repr__(self):
        return "@recursive_iterator(%s)" % (_coconut.repr(self.func),)
    def __reduce__(self):
        return (self.__class__, (self.func,))
    def __get__(self, obj, objtype=None):
        if obj is None:
            return self
        return _coconut.types.MethodType(self, obj)
class _coconut_FunctionMatchErrorContext:
    __slots__ = ("exc_class", "taken")
    threadlocal_ns = _coconut.threading.local()
    def __init__(self, exc_class):
        self.exc_class = exc_class
        self.taken = False
    @classmethod
    def get_contexts(cls):
        try:
            return cls.threadlocal_ns.contexts
        except _coconut.AttributeError:
            cls.threadlocal_ns.contexts = []
            return cls.threadlocal_ns.contexts
    def __enter__(self):
        self.get_contexts().append(self)
    def __exit__(self, type, value, traceback):
        self.get_contexts().pop()
def _coconut_get_function_match_error():
    try:
        ctx = _coconut_FunctionMatchErrorContext.get_contexts()[-1]
    except _coconut.IndexError:
        return _coconut_MatchError
    if ctx.taken:
        return _coconut_MatchError
    ctx.taken = True
    return ctx.exc_class
class _coconut_base_pattern_func(_coconut_base_hashable):
    if _coconut_sys.version_info < (3, 7):
        __slots__ = ("FunctionMatchError", "patterns", "__doc__", "__name__")
    else:
        __slots__ = ("FunctionMatchError", "patterns", "__doc__", "__name__", "__qualname__")
    _coconut_is_match = True
    def __init__(self, *funcs):
        self.FunctionMatchError = _coconut.type(_coconut_py_str("MatchError"), (_coconut_MatchError,), {})
        self.patterns = []
        self.__doc__ = None
        self.__name__ = None
        if _coconut_sys.version_info >= (3, 7):
            self.__qualname__ = None
        for func in funcs:
            self.add_pattern(func)
    def add_pattern(self, func):
        if _coconut.isinstance(func, _coconut_base_pattern_func):
            self.patterns += func.patterns
        else:
            self.patterns.append(func)
        self.__doc__ = _coconut.getattr(func, "__doc__", self.__doc__)
        self.__name__ = _coconut.getattr(func, "__name__", self.__name__)
        if _coconut_sys.version_info >= (3, 7):
            self.__qualname__ = _coconut.getattr(func, "__qualname__", self.__qualname__)
    def __call__(self, *args, **kwargs):
        for func in self.patterns[:-1]:
            try:
                with _coconut_FunctionMatchErrorContext(self.FunctionMatchError):
                    return func(*args, **kwargs)
            except self.FunctionMatchError:
                pass
        return self.patterns[-1](*args, **kwargs)
    def _coconut_tco_func(self, *args, **kwargs):
        for func in self.patterns[:-1]:
            try:
                with _coconut_FunctionMatchErrorContext(self.FunctionMatchError):
                    return func(*args, **kwargs)
            except self.FunctionMatchError:
                pass
        return _coconut_tail_call(self.patterns[-1], *args, **kwargs)
    def __repr__(self):
        return "addpattern(%s)(*%s)" % (_coconut.repr(self.patterns[0]), _coconut.repr(self.patterns[1:]))
    def __reduce__(self):
        return (self.__class__, _coconut.tuple(self.patterns))
    def __get__(self, obj, objtype=None):
        if obj is None:
            return self
        return _coconut.types.MethodType(self, obj)
def _coconut_mark_as_match(base_func):
    base_func._coconut_is_match = True
    return base_func
def addpattern(base_func, **kwargs):
    """Decorator to add a new case to a pattern-matching function,
    where the new case is checked last."""
    allow_any_func = kwargs.pop("allow_any_func", False)
    if not allow_any_func and not _coconut.getattr(base_func, "_coconut_is_match", False):
        _coconut.warnings.warn("Possible misuse of addpattern with non-pattern-matching function " + _coconut.repr(base_func) + " (pass allow_any_func=True to dismiss)", stacklevel=2)
    if kwargs:
        raise _coconut.TypeError("addpattern() got unexpected keyword arguments " + _coconut.repr(kwargs))
    return _coconut.functools.partial(_coconut_base_pattern_func, base_func)
_coconut_addpattern = addpattern
def prepattern(base_func, **kwargs):
    """DEPRECATED: use addpattern instead."""
    def pattern_prepender(func):
        return addpattern(func, **kwargs)(base_func)
    return pattern_prepender
class _coconut_partial(_coconut_base_hashable):
    __slots__ = ("func", "_argdict", "_arglen", "_stargs", "keywords")
    if hasattr(_coconut.functools.partial, "__doc__"):
        __doc__ = _coconut.functools.partial.__doc__
    def __init__(self, func, argdict, arglen, *args, **kwargs):
        self.func = func
        self._argdict = argdict
        self._arglen = arglen
        self._stargs = args
        self.keywords = kwargs
    def __reduce__(self):
        return (self.__class__, (self.func, self._argdict, self._arglen) + self._stargs, self.keywords)
    def __setstate__(self, keywords):
        self.keywords = keywords
    @property
    def args(self):
        return _coconut.tuple(self._argdict.get(i) for i in _coconut.range(self._arglen)) + self._stargs
    def __call__(self, *args, **kwargs):
        callargs = []
        argind = 0
        for i in _coconut.range(self._arglen):
            if i in self._argdict:
                callargs.append(self._argdict[i])
            elif argind >= _coconut.len(args):
                raise _coconut.TypeError("expected at least " + _coconut.str(self._arglen - _coconut.len(self._argdict)) + " argument(s) to " + _coconut.repr(self))
            else:
                callargs.append(args[argind])
                argind += 1
        callargs += self._stargs
        callargs += args[argind:]
        kwargs.update(self.keywords)
        return self.func(*callargs, **kwargs)
    def __repr__(self):
        args = []
        for i in _coconut.range(self._arglen):
            if i in self._argdict:
                args.append(_coconut.repr(self._argdict[i]))
            else:
                args.append("?")
        for arg in self._stargs:
            args.append(_coconut.repr(arg))
        return "%s$(%s)" % (_coconut.repr(self.func), ", ".join(args))
def consume(iterable, keep_last=0):
    """consume(iterable, keep_last) fully exhausts iterable and return the last keep_last elements."""
    return _coconut.collections.deque(iterable, maxlen=keep_last)
class starmap(_coconut_base_hashable, _coconut.itertools.starmap):
    __slots__ = ("func", "iter")
    if hasattr(_coconut.itertools.starmap, "__doc__"):
        __doc__ = _coconut.itertools.starmap.__doc__
    def __new__(cls, function, iterable):
        new_map = _coconut.itertools.starmap.__new__(cls, function, iterable)
        new_map.func = function
        new_map.iter = iterable
        return new_map
    def __getitem__(self, index):
        if _coconut.isinstance(index, _coconut.slice):
            return self.__class__(self.func, _coconut_igetitem(self.iter, index))
        return self.func(*_coconut_igetitem(self.iter, index))
    def __reversed__(self):
        return self.__class__(self.func, *_coconut_reversed(self.iter))
    def __len__(self):
        return _coconut.len(self.iter)
    def __repr__(self):
        return "starmap(%r, %r)" % (self.func, self.iter)
    def __reduce__(self):
        return (self.__class__, (self.func, self.iter))
    def __iter__(self):
        return _coconut.iter(_coconut.itertools.starmap(self.func, self.iter))
    def __fmap__(self, func):
        return self.__class__(_coconut_forward_compose(self.func, func), self.iter)
def makedata(data_type, *args):
    """Construct an object of the given data_type containing the given arguments."""
    if _coconut.hasattr(data_type, "_make") and _coconut.issubclass(data_type, _coconut.tuple):
        return data_type._make(args)
    if _coconut.issubclass(data_type, (_coconut.range, _coconut.abc.Iterator)):
        return args
    if _coconut.issubclass(data_type, _coconut.str):
        return "".join(args)
    return data_type(args)
def datamaker(data_type):
    """DEPRECATED: use makedata instead."""
    return _coconut.functools.partial(makedata, data_type)
def fmap(func, obj):
    """fmap(func, obj) creates a copy of obj with func applied to its contents.
    Override by defining obj.__fmap__(func). For numpy arrays, uses np.vectorize."""
    obj_fmap = _coconut.getattr(obj, "__fmap__", None)
    if obj_fmap is not None:
        try:
            result = obj_fmap(func)
        except _coconut.NotImplementedError:
            pass
        else:
            if result is not _coconut.NotImplemented:
                return result
    if obj.__class__.__module__ in ("numpy", "pandas"):
        from numpy import vectorize
        return vectorize(func)(obj)
    return _coconut_makedata(obj.__class__, *(_coconut_starmap(func, obj.items()) if _coconut.isinstance(obj, _coconut.abc.Mapping) else _coconut_map(func, obj)))
def memoize(maxsize=None, *args, **kwargs):
    """Decorator that memoizes a function,
    preventing it from being recomputed if it is called multiple times with the same arguments."""
    return _coconut.functools.lru_cache(maxsize, *args, **kwargs)
def _coconut_call_set_names(cls):
    if _coconut_sys.version_info < (3, 6):
        for k, v in _coconut.vars(cls).items():
            set_name = _coconut.getattr(v, "__set_name__", None)
            if set_name is not None:
                set_name(cls, k)
class override(_coconut_base_hashable):
    __slots__ = ("func",)
    def __init__(self, func):
        self.func = func
    def __get__(self, obj, objtype=None):
        if obj is None:
            return self.func
        return _coconut.types.MethodType(self.func, obj)
    def __set_name__(self, obj, name):
        if not _coconut.hasattr(_coconut.super(obj, obj), name):
            raise _coconut.RuntimeError(obj.__name__ + "." + name + " marked with @override but not overriding anything")
    def __reduce__(self):
        return (self.__class__, (self.func,))
def reveal_type(obj):
    """Special function to get MyPy to print the type of the given expression.
    At runtime, reveal_type is the identity function."""
    return obj
def reveal_locals():
    """Special function to get MyPy to print the type of the current locals.
    At runtime, reveal_locals always returns None."""
    pass
def _coconut_handle_cls_kwargs(**kwargs):
    metaclass = kwargs.pop("metaclass", None)
    if kwargs and metaclass is None:
        raise _coconut.TypeError("unexpected keyword argument(s) in class definition: %s" % (kwargs,))
    def coconut_handle_cls_kwargs_wrapper(cls):
        if metaclass is None:
            return cls
        orig_vars = cls.__dict__.copy()
        slots = orig_vars.get("__slots__")
        if slots is not None:
            if _coconut.isinstance(slots, str):
                slots = [slots]
            for slots_var in slots:
                orig_vars.pop(slots_var)
        orig_vars.pop("__dict__", None)
        orig_vars.pop("__weakref__", None)
        if _coconut.hasattr(cls, "__qualname__"):
            orig_vars["__qualname__"] = cls.__qualname__
        return metaclass(cls.__name__, cls.__bases__, orig_vars, **kwargs)
    return coconut_handle_cls_kwargs_wrapper
def _coconut_handle_cls_stargs(*args):
    temp_names = ["_coconut_base_cls_%s" % (i,) for i in _coconut.range(_coconut.len(args))]
    ns = _coconut.dict(_coconut.zip(temp_names, args))
    exec("class _coconut_cls_stargs_base(" + ", ".join(temp_names) + "): pass", ns)
    return ns["_coconut_cls_stargs_base"]
def _coconut_dict_merge(*dicts, **options):
    for_func = options.pop("for_func", False)
    assert not options, "error with internal Coconut function _coconut_dict_merge (you should report this at https://github.com/evhub/coconut/issues/new)"
    newdict = {}
    prevlen = 0
    for d in dicts:
        newdict.update(d)
        if for_func:
            if len(newdict) != prevlen + len(d):
                raise _coconut.TypeError("multiple values for the same keyword argument")
            prevlen = len(newdict)
    return newdict
_coconut_self_match_types = (bool, bytearray, bytes, dict, float, frozenset, int, list, set, str, tuple)
_coconut_MatchError, _coconut_count, _coconut_enumerate, _coconut_filter, _coconut_makedata, _coconut_map, _coconut_reiterable, _coconut_reversed, _coconut_starmap, _coconut_tee, _coconut_zip, TYPE_CHECKING, reduce, takewhile, dropwhile = MatchError, count, enumerate, filter, makedata, map, reiterable, reversed, starmap, tee, zip, False, _coconut.functools.reduce, _coconut.itertools.takewhile, _coconut.itertools.dropwhile



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# __coconut_hash__ = 0x37af8c9e

# Compiled with Coconut version 1.6.0 [Vocational Guidance Counsellor]

# Coconut Header: -------------------------------------------------------------

from __future__ import generator_stop
import sys as _coconut_sys, os as _coconut_os
_coconut_file_dir = _coconut_os.path.dirname(_coconut_os.path.abspath(__file__))
_coconut_cached_module = _coconut_sys.modules.get("__coconut__")
if _coconut_cached_module is not None and _coconut_os.path.dirname(_coconut_cached_module.__file__) != _coconut_file_dir:
    del _coconut_sys.modules["__coconut__"]
_coconut_sys.path.insert(0, _coconut_file_dir)
_coconut_module_name = _coconut_os.path.splitext(_coconut_os.path.basename(_coconut_file_dir))[0]
if _coconut_module_name and _coconut_module_name[0].isalpha() and all(c.isalpha() or c.isdigit() for c in _coconut_module_name) and "__init__.py" in _coconut_os.listdir(_coconut_file_dir):
    _coconut_full_module_name = str(_coconut_module_name + ".__coconut__")
    import __coconut__ as _coconut__coconut__
    _coconut__coconut__.__name__ = _coconut_full_module_name
    for _coconut_v in vars(_coconut__coconut__).values():
        if getattr(_coconut_v, "__module__", None) == "__coconut__":
            try:
                _coconut_v.__module__ = _coconut_full_module_name
            except AttributeError:
                _coconut_v_type = type(_coconut_v)
                if getattr(_coconut_v_type, "__module__", None) == "__coconut__":
                    _coconut_v_type.__module__ = _coconut_full_module_name
    _coconut_sys.modules[_coconut_full_module_name] = _coconut__coconut__
from __coconut__ import *
from __coconut__ import _coconut_call_set_names, _coconut, _coconut_MatchError, _coconut_igetitem, _coconut_base_compose, _coconut_forward_compose, _coconut_back_compose, _coconut_forward_star_compose, _coconut_back_star_compose, _coconut_forward_dubstar_compose, _coconut_back_dubstar_compose, _coconut_pipe, _coconut_star_pipe, _coconut_dubstar_pipe, _coconut_back_pipe, _coconut_back_star_pipe, _coconut_back_dubstar_pipe, _coconut_none_pipe, _coconut_none_star_pipe, _coconut_none_dubstar_pipe, _coconut_bool_and, _coconut_bool_or, _coconut_none_coalesce, _coconut_minus, _coconut_map, _coconut_partial, _coconut_get_function_match_error, _coconut_base_pattern_func, _coconut_addpattern, _coconut_sentinel, _coconut_assert, _coconut_mark_as_match, _coconut_reiterable, _coconut_self_match_types, _coconut_dict_merge, _coconut_exec
_coconut_sys.path.pop(0)

# Compiled Coconut: -----------------------------------------------------------

from omni_cv_rules.coconut.convert import to_imagedef  # from omni_cv_rules.coconut.convert import to_imagedef, ImageDef, Numpy, DataEdge
from omni_cv_rules.coconut.convert import ImageDef  # from omni_cv_rules.coconut.convert import to_imagedef, ImageDef, Numpy, DataEdge
from omni_cv_rules.coconut.convert import Numpy  # from omni_cv_rules.coconut.convert import to_imagedef, ImageDef, Numpy, DataEdge
from omni_cv_rules.coconut.convert import DataEdge  # from omni_cv_rules.coconut.convert import to_imagedef, ImageDef, Numpy, DataEdge



import io
import itertools
from functools import partial
from typing import List, Any

import matplotlib
import numpy as np
from loguru import logger
from matplotlib import pyplot as plt

from omni_converter.solver.rules import RuleEdge


def to_nrm_quiver_auto(nrm: "AutoData", auto_func) -> "AutoData":
    return to_nrm_quiver(nrm.to("numpy,float32,HWC,XYZ,-1_1"), auto_func)


def to_nrm_quiver(nrm: np.ndarray, auto_func) -> "AutoData":
    # fmt: numpy,float32,HWC,XYZ,-1_1
    # converting an image to nrm_quiver destructs semantic consistency. should I avoid adding this to rule?
    # I guess not.
    # we need 10 for 256 px so
    np_nrm = nrm  # == ("numpy,float32,HWC,XYZ,-1_1")
    h, w, c = np_nrm.shape
    figw = w / 256 * 10
    figh = h / 256 * 10
    logger.info(f"using fig size:{(figw, figh)}")
    fig, ax = plt.subplots(figsize=(figw, figh))
    quiver_sample_n = int(min(w, h) / 256 * 50)

    x, y = np.meshgrid(np.linspace(0, w - 1, quiver_sample_n), np.linspace(0, h - 1, quiver_sample_n))
    sampled = np_nrm[-y.astype(int), x.astype(int)]  # (n,n,3)
    quiver = ax.quiver(x, y, sampled[:, :, 0], sampled[:, :, 1])
    ax.tick_params(length=5)
    return plt_to_image(plt, auto_func)


def plt_to_image(plt: matplotlib.pyplot, auto_func):
    buf = io.BytesIO()
    plt.savefig(buf, format="png")
    buf.seek(0)
    img = plt.imread(buf)
    return auto_func("numpy,float32,HWC,RGBA,0_1", img)


def swap_channel_order(img: np.ndarray, src_idx, dst_idx):
    assert len(img.shape) == 4, "only support BHWC type of ndarray image"
    copied = np.copy(img)
    src = copied[:, :, :, src_idx]
    copied[:, :, :, src_idx] = copied[:, :, :, dst_idx]
    copied[:, :, :, dst_idx] = src
    return src


import re

CH_MATCHER = re.compile("[A-Z][a-z]*")


def extract_channels(ch_repr: str):
    return CH_MATCHER.findall(ch_repr)


def swap_element(items: list, i, j):
    items = items.copy()
    tmp = items[i]
    items[i] = items[j]
    items[j] = tmp
    return items


def rule_to_swap_channel_order(state):
    from pampy import match
    def _swap_rules(s: dict):
        channels = extract_channels(s["ch_rpr"])
        axis = list(range(len(channels)))
        res = []
        for src_i, dst_i in itertools.combinations(axis, 2):
            new_chs = swap_element(channels, src_i, dst_i)
            new_state = s.copy()
            new_state["ch_rpr"] = "".join(new_chs)
            res.append(RuleEdge(
                converter=partial(swap_channel_order, src_idx=src_i, dst_idx=dst_i),
                new_state=new_state,
                cost=1,
                name=f"swap channel {channels[src_i]} to {channels[dst_i]}"
            ))
        return res

    match(state,
          {"type": "numpy", "arrange": "BHWC"}, _swap_rules,
          Any, lambda state: None
          )


def rule_to_nrm_quiver(state) -> List[RuleEdge]:
    if state == "numpy,float32,HWC,XYZ,-1_1":
        return [RuleEdge(
            to_nrm_quiver,
            "nrm_quiver",
            1,
            "normal map to quiver plot"
        )]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# __coconut_hash__ = 0x4bbe2f00

# Compiled with Coconut version 1.6.0 [Vocational Guidance Counsellor]

# Coconut Header: -------------------------------------------------------------

from __future__ import generator_stop
import sys as _coconut_sys, os as _coconut_os
_coconut_file_dir = _coconut_os.path.dirname(_coconut_os.path.abspath(__file__))
_coconut_cached_module = _coconut_sys.modules.get("__coconut__")
if _coconut_cached_module is not None and _coconut_os.path.dirname(_coconut_cached_module.__file__) != _coconut_file_dir:
    del _coconut_sys.modules["__coconut__"]
_coconut_sys.path.insert(0, _coconut_file_dir)
_coconut_module_name = _coconut_os.path.splitext(_coconut_os.path.basename(_coconut_file_dir))[0]
if _coconut_module_name and _coconut_module_name[0].isalpha() and all(c.isalpha() or c.isdigit() for c in _coconut_module_name) and "__init__.py" in _coconut_os.listdir(_coconut_file_dir):
    _coconut_full_module_name = str(_coconut_module_name + ".__coconut__")
    import __coconut__ as _coconut__coconut__
    _coconut__coconut__.__name__ = _coconut_full_module_name
    for _coconut_v in vars(_coconut__coconut__).values():
        if getattr(_coconut_v, "__module__", None) == "__coconut__":
            try:
                _coconut_v.__module__ = _coconut_full_module_name
            except AttributeError:
                _coconut_v_type = type(_coconut_v)
                if getattr(_coconut_v_type, "__module__", None) == "__coconut__":
                    _coconut_v_type.__module__ = _coconut_full_module_name
    _coconut_sys.modules[_coconut_full_module_name] = _coconut__coconut__
from __coconut__ import *
from __coconut__ import _coconut_call_set_names, _coconut, _coconut_MatchError, _coconut_igetitem, _coconut_base_compose, _coconut_forward_compose, _coconut_back_compose, _coconut_forward_star_compose, _coconut_back_star_compose, _coconut_forward_dubstar_compose, _coconut_back_dubstar_compose, _coconut_pipe, _coconut_star_pipe, _coconut_dubstar_pipe, _coconut_back_pipe, _coconut_back_star_pipe, _coconut_back_dubstar_pipe, _coconut_none_pipe, _coconut_none_star_pipe, _coconut_none_dubstar_pipe, _coconut_bool_and, _coconut_bool_or, _coconut_none_coalesce, _coconut_minus, _coconut_map, _coconut_partial, _coconut_get_function_match_error, _coconut_base_pattern_func, _coconut_addpattern, _coconut_sentinel, _coconut_assert, _coconut_mark_as_match, _coconut_reiterable, _coconut_self_match_types, _coconut_dict_merge, _coconut_exec
_coconut_sys.path.pop(0)

# Compiled Coconut: -----------------------------------------------------------

from omni_cv_rules.coconut.omni_converter import AutoList  # from omni_cv_rules.coconut.omni_converter import AutoList


def intra_list_conversions(state, neighbors):  # def intra_list_conversions(state,neighbors):
    _coconut_case_match_to_0 = state  #     case state:
    _coconut_case_match_check_0 = False  #     case state:
    if (_coconut.isinstance(_coconut_case_match_to_0, AutoList)) and (_coconut.len(_coconut_case_match_to_0) == 1) and (_coconut.isinstance(_coconut_case_match_to_0[0], AutoList)) and (_coconut.len(_coconut_case_match_to_0[0]) == 1):  #     case state:
        _coconut_case_match_check_0 = True  #     case state:
    if _coconut_case_match_check_0:  #     case state:
        return ([])  # dont dive into nested list too much  #             return [] # dont dive into nested list too much
    if not _coconut_case_match_check_0:  #         match AutoList(es):
        _coconut_match_set_name_es = _coconut_sentinel  #         match AutoList(es):
        if (_coconut.isinstance(_coconut_case_match_to_0, AutoList)) and (_coconut.len(_coconut_case_match_to_0) == 1):  #         match AutoList(es):
            _coconut_match_set_name_es = _coconut_case_match_to_0[0]  #         match AutoList(es):
            _coconut_case_match_check_0 = True  #         match AutoList(es):
        if _coconut_case_match_check_0:  #         match AutoList(es):
            if _coconut_match_set_name_es is not _coconut_sentinel:  #         match AutoList(es):
                es = _coconut_case_match_to_0[0]  #         match AutoList(es):
        if _coconut_case_match_check_0:  #         match AutoList(es):
            return ([(lambda f, new_state, cost, name: (lambda items: [f(i) for i in items], AutoList(new_state), "[{_coconut_format_0}]".format(_coconut_format_0=(name)), cost + 1))(re.converter, re.new_format, re.cost, re.name) for re in neighbors(es)])  #             return [((f,new_state,cost,name)->(



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['omni_cv_rules',
 'omni_cv_rules.coconut',
 'omni_cv_rules.place_holder',
 'omni_cv_rules.py_310',
 'omni_cv_rules.py_38']

package_data = \
{'': ['*']}

install_requires = \
['Pillow',
 'attrs',
 'bqplot',
 'frozendict',
 'ipywidgets',
 'numpy',
 'pandas',
 'pprintpp',
 'py-omni-converter']

setup_kwargs = {
    'name': 'omni-cv-rules',
    'version': '0.1.13',
    'description': '',
    'long_description': None,
    'author': 'proboscis',
    'author_email': 'nameissoap@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
